var createNodeKey = require('./create-node-key').createNodeKey;
var AstNodeTypes = require("./ast-node-types.js");
var SlJsInfra = require("sl-js-infra").SlJsInfra;

/*
When working in React, components are being generated by the "_createClass" method of the framework.
As a result, we have an issue when we try to map it back to source maps.
The method make sure to verify if a component is a React component by going through the parent nodes
fpr a specific node, and verifying that they have the same pattern as the expected code.

React Code
================== START CODE ============
 import React, { Component } from 'react';
 import logo from './logo.svg';
 import './App.css';

 class App extends Component {
    someFunc(){
        console.log("hezi func")
    }
 }

 export default App;
 ================== END CODE ============

This code generates the following code
 ================== START CODE ============
 _createClass(App, [{
    key: 'someFunc',
    value: function someFunc() {
        console.log("Some func is invoked");
    }
}]);
 ================== END CODE ============

So basically  we are searching for the "value" node, the array which contains it and the "_createClass" method.
* */
function isReact(node, nodeToParent){
    if (node.type != "FunctionExpression")
        return false;
    var key = createNodeKey(node);
    var valueNode = nodeToParent[key];
    if (valueNode == null)
        return false;
    if (valueNode.key == null || (valueNode.key !=null && valueNode.key.name != "value"))
        return false; //That's note the 'value' node. Return.
    key = createNodeKey(valueNode);
    parent = nodeToParent[key];
    if (parent == null)
        return false;
    var objectNode = parent;
    key = createNodeKey(objectNode);
    arrayNode = nodeToParent[key];
    if (arrayNode == null || arrayNode.type != "ArrayExpression")
        return false;
    key = createNodeKey(arrayNode);
    var createClassNode = nodeToParent[key];
    if (createClassNode == null || createClassNode.callee == null){
        return false;
    }
    var result = createClassNode.callee.name == "_createClass";
    return result;
}

/**

 * */
function guessMethodLocation(node, nodeToParent) {
    if (isReact(node, nodeToParent) && node.type == AstNodeTypes.FunctionExpression) {
        return node.body.loc;
    }

    /* According to this link, Istanbul uses the "node.id.loc" for FunctionDeclaration and FunctionExpression.
     * https://github.com/istanbuljs/istanbuljs/blob/cbd1c1473da0dc34051c332170544529d59eda03/packages/istanbul-lib-instrument/src/visitor.js#L216
     */
    if (!SlJsInfra.SlEnvVars.isUseIstanbul() && shouldUseLocFromID(node) && node.id && node.id.loc) {
        return node.id.loc;
    }


    /*
        In typescript, in case of below code, diferrent location recieved from source maps and babel parser.
        export function foo(){

        }
        Source maps: column 0
        Babel parser: cloumn 7 (the 'function' woord.
        Here we fix the location to be column 0 (export statement location)
     */
    if(node.type === "FunctionDeclaration"){
        var key = createNodeKey(node);
        var parentNode = nodeToParent[key];
        if(parentNode.type == "ExportDefaultDeclaration" || parentNode.type == "ExportNamedDeclaration") {
            return parentNode.loc
        }
    }
    return node.loc;
}

function shouldUseLocFromID(node) {
    return node.type === AstNodeTypes.FunctionDeclaration || node.type === AstNodeTypes.FunctionExpression;
}

module.exports.guessMethodLocation = guessMethodLocation;